#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <string.h>
#include "pmandel.h"

#define  MASTER  0
#define  TAG 17
#define  REQUEST 18
#define  BCAST 19
int  np, myid;
Parameter_t Param;
MPI_Datatype MPI_Param, MPI_Complex;
MPE_Color *colorArray;

void createDatatypes()
{
	MPI_Aint param_disp[6];
	int param_blocks[] = {4, 2};
	MPI_Datatype param_type[2];

	/* Initialise Complex */
	MPI_Type_vector(2, 1, 1, MPI_DOUBLE, &MPI_Complex);
	MPI_Type_commit(&MPI_Complex);

	/* Initialise Pack */
	param_type[0] = MPI_INT;
	param_type[1] = MPI_Complex;
	MPI_Address(&Param.CurrentLine, &param_disp[0]);
	MPI_Address(&Param.cStart, &param_disp[1]);
	param_disp[1] -= param_disp[0];
	param_disp[0] -= param_disp[0];
	MPI_Type_struct(2, param_blocks, param_disp, param_type, &MPI_Param);
	MPI_Type_commit(&MPI_Param);
}

void freeDatatypes()
{
	MPI_Type_free(&MPI_Param);
	MPI_Type_free(&MPI_Complex);
}

void exchangeParam()
{
	static int mode = 0;
	static unsigned char buf[sizeof(Parameter_t)];
	int bytes, pos = 0;
	MPI_Status status;

	bytes = sizeof(Parameter_t);

	if (mode == 0) /* toogle in every round the mode to solve both tasks */
	{
		MPI_Bcast(&Param, 1, MPI_Param, MASTER, MPI_COMM_WORLD);
		mode = 1;
	} else {
		if (myid == MASTER) {
			MPI_Pack(&Param.CurrentLine, 4, MPI_INT, buf, bytes, &pos, MPI_COMM_WORLD);
			MPI_Pack(&Param.cStart, 4, MPI_DOUBLE, buf, bytes, &pos, MPI_COMM_WORLD);
			MPI_Bcast(&buf, bytes, MPI_PACKED, MASTER, MPI_COMM_WORLD);
		} else {
			MPI_Bcast(&buf, bytes, MPI_PACKED, MASTER, MPI_COMM_WORLD);
			MPI_Unpack(buf, bytes, &pos, &Param.CurrentLine, 4, MPI_INT, MPI_COMM_WORLD);
			MPI_Unpack(buf, bytes, &pos, &Param.cStart, 4, MPI_DOUBLE, MPI_COMM_WORLD);
		}
		mode = 0;
	}
}

int main(int argc, char **argv)
{
	int   myWindowOpened, allWindowsOpened;
	Winspecs  winspecs;
	MPI_Status  status;
	int  i, j, k;
	char  hostname[80];

	MPI_Init(&argc, &argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &myid);
	MPI_Comm_size(MPI_COMM_WORLD, &np);

	gethostname(hostname, 80);
	printf("Hi from Task %d on %s\n", myid, hostname);
	fflush(stdout);

	myWindowOpened = InitGraphics(&winspecs);
	MPI_Allreduce(&myWindowOpened, &allWindowsOpened, 1, MPI_INT, MPI_LAND, MPI_COMM_WORLD);

	if (allWindowsOpened != 0) {
		MakeColors(&winspecs);

		InitParams(&Param, &winspecs);
		createDatatypes();

		do /* render until single click */
		{
			Param.CurrentLine = 0; /* reset current line */

			if (myid == MASTER) {
				Master(); /* let master control rendering */
				DragZoom(&winspecs); /* allow user input after redering */
			} else {
				Slave(&winspecs); /* let slaves render data from master */
			}

			exchangeParam();
		} while (Param.CurrentLine != -1);

		printf("Task %d on %s ready\n", myid, hostname);
		MPI_Barrier(MPI_COMM_WORLD);
		MPE_Close_graphics(&winspecs.graph);
	} else {
		if (myid == MASTER) {
			fprintf(stderr, "One or more processes could not connect\n");
			fprintf(stderr, "to the display.  Exiting.\n\n");
		}
		if (myWindowOpened)
			MPE_Close_graphics(&winspecs.graph);
	}

	freeDatatypes();
	MPI_Finalize();
}

int DragZoom(Winspecs *winspecs)
{
	int xLeft, yTop, xRight, yBottom, i, button;

	MPE_Get_drag_region(winspecs->graph, 1, MPE_DRAG_SQUARE, &xLeft, &yTop, &xRight, &yBottom);
	Param.cStart.real += min(xLeft, xRight) * Param.cStep.real;
	Param.cStart.imag += min(yTop, yBottom) * Param.cStep.imag;
	Param.cStep.real *= (float) + abs(xRight - xLeft) / Param.PicWidth;
	Param.cStep.imag *= (float) + abs(yBottom - yTop) / Param.PicHeight;

	if (abs(xLeft - xRight) < 5) {
		Param.CurrentLine = -1;
	}
	return 1;
}

void InitParams(Parameter_t *Param, Winspecs *winspecs)
{
	Param->CurrentLine = 0;
	Param->nRows       = 20;
	Param->PicWidth    = winspecs->width;
	Param->PicHeight   = winspecs->height;
	Param->cStart.real = -2.3;
	Param->cStart.imag = -1.5;
	Param->cStep.real  = 3.5 / Param->PicWidth;
	Param->cStep.imag  = 3.0 / Param->PicHeight;
}

/* Emit some line rows to slaves and be sure that every slave stops */
int Master()
{
	MPI_Status status;
	int i, slave_id, slaves;

	/* Emit data do start threads*/
	slaves = 0;
	for (i = 1; i < np; i++) {
		if (Param.CurrentLine < Param.PicHeight) /* We just use as much nodes as we can and NOT MORE */
		{
			MPI_Send(&Param.CurrentLine, 1, MPI_INT, i, TAG, MPI_COMM_WORLD); /* use ssend because I dont know if send is buffered */
			Param.CurrentLine += Param.nRows;
			slaves++;
		} else /* These nodes aren't needed -> just let them quit there loop */
		{
			MPI_Send(&Param.CurrentLine, 1, MPI_INT, i, TAG, MPI_COMM_WORLD);
		}
	}

	/* Emit data for ready slaves until picture is rendered */
	while (Param.CurrentLine < Param.PicHeight) {
		MPI_Recv(&slave_id, 1, MPI_INT, MPI_ANY_SOURCE, REQUEST, MPI_COMM_WORLD, &status);
		MPI_Send(&Param.CurrentLine, 1, MPI_INT, slave_id, TAG, MPI_COMM_WORLD); /* use ssend because I dont know if send is buffered */

		Param.CurrentLine += Param.nRows;
	}

	/* Emit "stop" data to all remaining slaves - one has stopped by is one (that with the last segment) */
	for (i = 1; i < slaves; i++) {
		MPI_Recv(&slave_id, 1, MPI_INT, MPI_ANY_SOURCE, REQUEST, MPI_COMM_WORLD, &status);
		MPI_Send(&Param.CurrentLine, 1, MPI_INT, slave_id, TAG, MPI_COMM_WORLD);
	}

	return 0;
}

int Slave(Winspecs *winspecs)
{
	MPI_Status status;
	int line;
	MPE_Point *ImageData;

	ImageData = (MPE_Point*)malloc(Param.nRows * Param.PicWidth * sizeof(MPE_Point)); /* create buffer for the pixel(color, x, y) for later output */

	do /* receive data, calculate date (+output) and send request for more when we havent reached the last row */
	{
		MPI_Recv(&line, 1, MPI_INT, MASTER, TAG, MPI_COMM_WORLD, &status);

		Param.CurrentLine = line;
		calculate(ImageData, winspecs);

		MPE_Draw_points(winspecs->graph, ImageData, Param.nRows*Param.PicWidth);
		MPE_Update(winspecs->graph);

		/* Only send "I want more" request, when we havent finished everything */
		if (Param.CurrentLine < Param.PicHeight) {
			MPI_Send(&myid, 1, MPI_INT, MASTER, REQUEST, MPI_COMM_WORLD);
		}

	} while (Param.CurrentLine < Param.PicHeight);

	return 0;
}
