#include <stdio.h>
#include "utils.h"

#ifndef _STDLIB_H
#include <stdlib.h>
#endif

#ifndef _STRING_H
#include <string.h>
#endif

extern int np, myid;
extern Parameter_t Param;

int InitGraphics(Winspecs *winspecs)
{
	int myWindowOpened;
	char display[80], *displayvar; //hostname[80];

	if (!myid) {
		// gethostname(hostname, 79);
		// sprintf(display,"%s:0.0",hostname);
		displayvar = getenv("DISPLAY");
		if (displayvar == NULL) {
			printf("Die Umgebungsvariable $DISPLAY ist nicht gesetzt!\n");
			snprintf(display, 3, "no");
		} else
			snprintf(display, strlen(displayvar), "%s", displayvar);

	}
	MPI_Bcast(display, 80, MPI_CHAR, 0, MPI_COMM_WORLD);

	if (!strcmp(display, "no")) {
		MPI_Finalize();
		printf("Das Programm wird beendet!\n");
		exit(1);
	}

	winspecs->numColors = 238;
	winspecs->xpos      = 256;
	winspecs->ypos      = 256;
	winspecs->width     = 600;
	winspecs->height    = 600;
	winspecs->bw        = 0;

	myWindowOpened = (MPE_Open_graphics(& (winspecs->graph), MPI_COMM_WORLD,
	                                    display,
	                                    winspecs->xpos, winspecs->ypos,
	                                    winspecs->width, winspecs->height, 1)
	                  == MPE_SUCCESS);

	return myWindowOpened;
}


int calculate(MPE_Point *ImageData, Winspecs *winspecs)
{
	register int x, iteration, i;
	register double cReal, cImag, zReal, zImag, zSqrReal, zSqrImag;
	MPE_Point *PicData;

	PicData = ImageData ;
	for (i = 0; i < Param.nRows && Param.CurrentLine < Param.PicHeight; i++, Param.CurrentLine++) {
		cImag = Param.cStart.imag + Param.CurrentLine * Param.cStep.imag;
		for (x = 0, cReal = Param.cStart.real; x < Param.PicWidth; x++, cReal += Param.cStep.real) {
			PicData->x = x;
			PicData->y = Param.CurrentLine;
			zReal = cReal; zImag = cImag; iteration = 1;
			do {
				zSqrReal = zReal * zReal;
				zSqrImag = zImag * zImag;
				/* zImag = 2*zReal*zImag + cImag;
				zReal = zSqrReal - zSqrImag + cReal; */
				zImag = 2 * zReal * zImag - 0.52542;
				zReal = zSqrReal - zSqrImag - 0.51145f;
			} while (iteration++ < MAXITER && zSqrReal + zSqrImag < DIVERGENCE);
			PicData->c = iteration < MAXITER ? iteration : 1;
			PicData++;
		}
	}
}

int MakeColors(Winspecs *winspecs)
{
	double Red, Green, Blue, ColorStep;
	int red, green, blue, i, ColorRange, nRanges;

	if (!winspecs->bw) {
		nRanges = 4;
		ColorRange = winspecs->numColors / nRanges;
		ColorStep = 200.0 / ((double) winspecs->numColors / (double) nRanges);
		winspecs->colorArray = (MPE_Color *) malloc(winspecs->numColors *
		                       sizeof(MPE_Color));
		red = 255; green = 53; blue = 53;
		Red = 255; Green = 53; Blue = 53;
		red   <<= 8;
		blue  <<= 8;
		for (i = 0; i < ColorRange; ++i) {
			green <<= 8;
			MPE_Add_RGB_color(winspecs->graph,
			                  red, green, blue, winspecs->colorArray + i);
			Green += ColorStep;
			green = (int) Green;
		}

		red = 255; green = 255; blue = 53;
		Red = 255; Green = 255; Blue = 53;
		green <<= 8;
		blue  <<= 8;
		for (; i < 2*ColorRange; ++i) {
			red   <<= 8;
			MPE_Add_RGB_color(winspecs->graph,
			                  red, green, blue, winspecs->colorArray + i);
			Red -= ColorStep;
			red = (int) Red;
		}

		red = 53; green = 255; blue = 53;
		Red = 53; Green = 255; Blue = 53;
		green <<= 8;
		red   <<= 8;
		for (; i < 3*ColorRange; ++i) {
			blue <<= 8;
			MPE_Add_RGB_color(winspecs->graph,
			                  red, green, blue, winspecs->colorArray + i);
			Blue += ColorStep;
			blue  = (int) Blue;
		}

		red = 53; green = 255; blue = 255;
		Red = 53; Green = 255; Blue = 255;
		blue <<= 8;
		red  <<= 8;
		for (; i < winspecs->numColors; ++i) {
			green <<= 8;
			MPE_Add_RGB_color(winspecs->graph,
			                  red, green, blue, winspecs->colorArray + i);
			Green -= ColorStep;
			green  = (int) Green;
		}
	}
	return 1;
}
